﻿# Electronic Annex 2: Setting of the model to individual test-topology configurations,
# running the AC-power-flow analysis solution using Gauss-Seidel method,
# Newton-Raphson method, or Newton-Raphson method with Iwamoto multipliers, and 
# inserting the AC-power-flow analysis computation times to the result matrix
# Jan Vysocky, Ladislav Foltyn, Dejan Brkic, Renáta Praksova, Pavel Praks

import pandapower
import numpy
from test_grid import make_test_grid
from scipy.io import loadmat
import pandas as pd
import pp2scipy
# Run this script three times:
# - at the first run, we will set the script to use the Gauss-Seidel method when solving power-flow analysis
# - on the second run, we will set the script to use the Newton-Raphson method when solving power-flow analysis
# - on the third run, we set the script to use the Newton-Raphson method with Iwamoto coefficients when solving
#   power-flow analysis
# … Power-flow analysis solution method is set through parameter "num_method" below:
#   1 - "gs" for Gauss-Seidel method
#   2 -  "nr" for Newton-Raphson method
#   3 - "iwamoto_nr" for Newton-Raphson method with Iwamoto coefficients


# Select number of the method
num_method = 2

if num_method == 1:
    method = 'gs'
elif num_method == 2:
    method = 'nr'
elif num_method == 3:
    method = 'iwamoto_nr'
else:
    print('Wrong number of the method! Gauss-Seidel method is set.')
    method = 'gs'

# Create a test grid from .py file
DS_model = make_test_grid()

# Do the initializing steady-state-power-flow analysis solution for the created pandapower network model
pandapower.runpp(net=DS_model, algorithm=method, calculate_voltage_angles='auto', init='auto', max_iteration=10,
                 tolerance_mva=1e-06, trafo_model='t', trafo_loading='current', enforce_q_lims=False,
                 check_connectivity=True, voltage_depend_loads=False, consider_line_temperature=False)

DS_model_switch_basic = DS_model.switch.copy()
DS_model["calculation time"] = 0

# create matrix measured_time_power_flow - its parameters are 100 rows and 90 columns - individual local rows will
# contain calcualtion time values for individual network combinations and individual local columns will contain
# calcualtion time values for individual numebers of switched-off power-line sections
measured_time_power_flow = numpy.zeros((100, 150), dtype=float)

# Load data saying which power-line sections are switched off in individual combinations, from the file …
# switched_off_power_line_sections_combinations.mat contains data of all combinations for all numbers of switched off
# power-line sections
loaded_data = loadmat('switched_off_power_line_sections_combinations.mat')
all_switchedoff_powerline_sections_combinations = loaded_data['all_switched-off_power-line_sections_combinations']

# Individual cells of this data structure contain all network combinations data for certain number of power-line
# sections switched-off (the data is sorted from 1 power-line section switched-off to 150 power-line sections
# switched-off here) … go from cell (0) to cell (149):

for n in range(150):
    # n = (number of power-line sections switched-off) - 1
    print('Number of power-line sections switched-off:', n+1)
    # ---------- Load the content of the n cell into the data structure all_combinations_for_n_switchedoff
    all_combinations_for_n_switchedoff = all_switchedoff_powerline_sections_combinations[n][0]
    # - the content of each cell is a matrix of combinations for a given number of power-line sections switched-off
    # ---- the number of combinations is always 100
    # --------- Take individual combinations from line (0) to line (99):
    for combination_ID in range(1):
        if (combination_ID % 10) == 0:
            print(combination_ID)
        DS_model.switch = DS_model_switch_basic.copy()
        current_combination = all_combinations_for_n_switchedoff[combination_ID]
        # Apply the given power-line section combination to the network model, i.e., switch off those switches whose
        # IDs = (power-line section index loaded from the vector current_combination) + (1e6) for ID
        # of switch 1… + (2e6) for ID of switch 2
        # - go from cell (0) to cell (length(current_combination)) in current_combination:
        for switchedoff_powerline_section_ID in range(len(current_combination)):
            switchedoff_powerline_section_index = int(current_combination[switchedoff_powerline_section_ID]+1e6)
            # Switch off both switches of the given power-line section in the network model
            switch_1_ID = int(switchedoff_powerline_section_index+1e6)
            switch_2_ID = int(switchedoff_powerline_section_index+2e6)
            DS_model.switch['closed'][switch_1_ID] = False
            DS_model.switch['closed'][switch_2_ID] = False
        # The network model is now set to the desired configuration
        # - Run power-flow analysis solution… during this process, the calulation time is measured and this value is
        #   saved in DS_model ["calculation time"]
        # - Take the current calculation time value from DS_model["calculation time"] and put it to the “combination_ID”
        #   row and “n” column of the matrix zmerene_casy_power_flow
        pandapower.runpp(net=DS_model, algorithm=method, calculate_voltage_angles='auto', init='auto', max_iteration=100,
                         tolerance_mva=1e-06, trafo_model='t', trafo_loading='current', enforce_q_lims=False,
                         check_connectivity=True, voltage_depend_loads=False, consider_line_temperature=False)
        measured_time_power_flow[combination_ID][n] = DS_model["calculation time"]


