# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2020 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

# Electronic Annex 3: The developed interface to pandapower solvers of the SciPy package  of the paper:
# Steady-state analysis of electrical networks in pandapower software: Computational performances of Newton-Raphson,
# Newton-Raphson with Iwamoto multiplier #and Gauss-Seidel methods
# Jan Vysocký, Ladislav Foltyn, Dejan Brkić, Renáta Praksová, Pavel Praks
"""Solves the power flow using SciPy methods.
"""
import json
from scipy import optimize

from numpy import angle, exp, linalg, conj, r_, Inf, arange, zeros, max, zeros_like, array
import numpy as np
from scipy.linalg import solve, inv
import time

from pandapower.auxiliary import _check_bus_index_and_print_warning_if_high, \
    _check_gen_index_and_print_warning_if_high, _init_runpp_options, \
    _internal_stored, _add_auxiliary_elements
from pandapower.results import init_results, verify_results
from pandapower.powerflow import _recycled_powerflow
from pandapower.run import _passed_runpp_parameters
from pandapower.pf.run_newton_raphson_pf import ppci_to_pfsoln, _get_numba_functions, _get_Y_bus, _get_Sbus, _store_internal
from pandapower.pf.ppci_variables import _get_pf_variables_from_ppci, _store_results_from_pf_in_ppci
from pandapower.pf.run_dc_pf import _run_dc_pf
from pandapower.pd2ppc import _pd2ppc
from pandapower.powerflow import _ppci_to_net


def _run_scipy_ac_pf(ppci, options):
    makeYbus, pfsoln = _get_numba_functions(ppci, options)

    baseMVA, bus, gen, branch, ref, pv, pq, _, _, V0, ref_gens = _get_pf_variables_from_ppci(ppci)

    ppci, Ybus, Yf, Yt = _get_Y_bus(ppci, options, makeYbus, baseMVA, bus, branch)

    # compute complex bus power injections [generation - load]
    Sbus = _get_Sbus(ppci, options["recycle"])

    # run power flow
    V, success, iterations, J, Vm_it, Va_it = _scipypf(Ybus, Sbus, V0, pv, pq, options)

    # keep "internal" variables in  memory / net["_ppc"]["internal"] -> needed for recycle.
    ppci = _store_internal(ppci, {"J": J, "Vm_it": Vm_it, "Va_it": Va_it, "bus": bus, "gen": gen, "branch": branch,
                                  "baseMVA": baseMVA, "V": V, "pv": pv, "pq": pq, "ref": ref, "Sbus": Sbus,
                                  "ref_gens": ref_gens, "Ybus": Ybus, "Yf": Yf, "Yt": Yt})

    return ppci, success, iterations


def _run_scipy_pf(ppci, options):
    if isinstance(options["init_va_degree"], str) and options["init_va_degree"] == "dc":
        ppci = _run_dc_pf(ppci)

    start = time.process_time()
    ppci, success, iterations = _run_scipy_ac_pf(ppci, options)
    # update data matrices with solution store in ppci
    bus, gen, branch = ppci_to_pfsoln(ppci, options)
    et = time.process_time() - start
    ppci = _store_results_from_pf_in_ppci(ppci, bus, gen, branch, success, iterations, et)
    return ppci


def _run_scipy_pf_algorithm(ppci, options, **kwargs):
    ac = options["ac"]

    if ac:
        result = _run_scipy_pf(ppci, options)
    else:
        raise Exception("DC powerflow is not involved!")

    return result


def _powerflow_scipy(net, **kwargs):
    """
    Gets called by runpp_scipy.
    """

    # get infos from options
    ac = net["_options"]["ac"]

    net["converged"] = False
    net["OPF_converged"] = False
    _add_auxiliary_elements(net)

    if not ac or net["_options"]["init_results"]:
        verify_results(net)
    else:
        init_results(net)


    # clear lookups
    net._pd2ppc_lookups = {"bus": array([], dtype=int), "ext_grid": array([], dtype=int),
                           "gen": array([], dtype=int), "branch": array([], dtype=int)}

    # convert pandapower net to ppc
    ppc, ppci = _pd2ppc(net)

    # store variables
    net["_ppc"] = ppc

    if not "VERBOSE" in kwargs:
        kwargs["VERBOSE"] = 0

    # ----- run the powerflow -----
    result = _run_scipy_pf_algorithm(ppci, net["_options"], **kwargs)
    # read the results (=ppci with results) to net
    _ppci_to_net(result, net)


def runpp_scipy(net, algorithm='nr', calculate_voltage_angles="auto", init="auto",
                max_iteration="auto", tolerance_mva=1e-8, trafo_model="t",
                trafo_loading="current", enforce_q_lims=False, check_connectivity=True,
                voltage_depend_loads=True, consider_line_temperature=False,
                run_control=False, **kwargs):
    """
        Description could be find in the original pandapower runpp function, but scipy algorithm are used here.

        Only AC power flow is involved without qlims enforced.
    """

    # if dict 'user_pf_options' is present in net, these options overrule the net.__internal_options
    # except for parameters that are passed by user
    recycle = kwargs.get("recycle", None)
    if isinstance(recycle, dict) and _internal_stored(net):
        _recycled_powerflow(net, **kwargs)
        return

    if run_control and net.controller.in_service.any():
        from pandapower.control import run_control
        parameters = {**locals(), **kwargs}
        # disable run control for inner loop to avoid infinite loop
        parameters["run_control"] = False
        run_control(**parameters)
    else:
        passed_parameters = _passed_runpp_parameters(locals())
        _init_runpp_options(net, algorithm=algorithm, calculate_voltage_angles=calculate_voltage_angles,
                            init=init, max_iteration=max_iteration, tolerance_mva=tolerance_mva,
                            trafo_model=trafo_model, trafo_loading=trafo_loading,
                            enforce_q_lims=enforce_q_lims, check_connectivity=check_connectivity,
                            voltage_depend_loads=voltage_depend_loads,
                            consider_line_temperature=consider_line_temperature,
                            passed_parameters=passed_parameters, **kwargs)
        _check_bus_index_and_print_warning_if_high(net)
        _check_gen_index_and_print_warning_if_high(net)
        _powerflow_scipy(net, **kwargs)


def Fx(VaVm):
    # evalute F(x)
    fromjson = json.load(open('fxjson.json'))
    slackids = fromjson['slackids']
    Yreal = np.array(fromjson['Yreal'])
    Yimag = np.array(fromjson['Yimag'])
    Sreal = np.array(fromjson['Sreal'])
    Simag = np.array(fromjson['Simag'])
    idxs = np.array(fromjson['pvpq'])

    hid = len(idxs)
    Va = np.array(VaVm[:hid])
    Vm = np.array(VaVm[hid:])

    F = np.zeros(2*hid, float)
    for i in range(hid):
        sum_p = 0
        sum_q = 0
        for j in range(hid):
            sum_p += Vm[i] * Vm[j] * (Yreal[idxs[i], idxs[j]] * np.cos(Va[i] - Va[j]) + Yimag[idxs[i], idxs[j]] * np.sin(Va[i] - Va[j]))
            sum_q += Vm[i] * Vm[j] * (Yreal[idxs[i], idxs[j]] * np.sin(Va[i] - Va[j]) - Yimag[idxs[i], idxs[j]] * np.cos(Va[i] - Va[j]))

        for j in slackids:
            sum_p += Vm[i] * (Yreal[idxs[i], j] * np.cos(Va[i]) + Yimag[idxs[i], j] * np.sin(Va[i]))
            sum_q += Vm[i] * (Yreal[idxs[i], j] * np.sin(Va[i]) - Yimag[idxs[i], j] * np.cos(Va[i]))

        F[i] = sum_p - Sreal[idxs[i]]
        F[i+hid] = sum_q - Simag[idxs[i]]
    #print('Error of ' + fromjson['algorithm'] + ': ' + str(linalg.norm(F, Inf)))
    return F


def fxjac(VaVm):
    fromjson = json.load(open('fxjson.json'))
    slackids = fromjson['slackids']
    Yreal = np.array(fromjson['Yreal'])
    Yimag = np.array(fromjson['Yimag'])
    idxs = np.array(fromjson['pvpq'])
    hid = len(idxs)
    Va = np.array(VaVm[:hid])
    Vm = np.array(VaVm[hid:])

    jac = np.zeros((2 * hid, 2 * hid), float)

    for i in range(hid):
        # derivative of P equation by variables
        sum_a = 0
        # F by Va[i]
        for j in range(hid):
            if i != j:
                sum_a += Vm[i] * Vm[j] * (-Yreal[idxs[i], idxs[j]] * np.sin(Va[i] - Va[j]) + Yimag[idxs[i], idxs[j]] * np.cos(Va[i] - Va[j]))
                jac[i, j] = Vm[i] * Vm[j] * (Yreal[idxs[i], idxs[j]] * np.sin(Va[i] - Va[j]) - Yimag[idxs[i], idxs[j]] * np.cos(Va[i] - Va[j]))
            else:
                sum_a += 0

        for j in slackids:
            sum_a += Vm[i] * (-Yreal[idxs[i], j] * np.sin(Va[i]) + Yimag[idxs[i], j] * np.cos(Va[i]))
        jac[i, i] = sum_a

        sum_v = 0
        # F by Vm[i]
        for j in range(hid):
            if i != j:
                sum_v += Vm[j] * (Yreal[idxs[i], idxs[j]] * np.cos(Va[i] - Va[j]) + Yimag[idxs[i], idxs[j]] * np.sin(Va[i] - Va[j]))
                jac[i, j+hid] = Vm[i] * (Yreal[idxs[i], idxs[j]] * np.cos(Va[i] - Va[j]) + Yimag[idxs[i], idxs[j]] * np.sin(Va[i] - Va[j]))
            else:
                sum_v += 2*Vm[i] * Yreal[idxs[i], idxs[j]]

        for j in slackids:
            sum_v += (Yreal[idxs[i], j] * np.cos(Va[i]) + Yimag[idxs[i], j] * np.sin(Va[i]))
        jac[i, i+hid] = sum_v

        # Derivative of Q by variables
        sum_a = 0
        # F by Va[i]
        for j in range(hid):
            if i != j:
                sum_a += Vm[i] * Vm[j] * (Yreal[idxs[i], idxs[j]] * np.cos(Va[i] - Va[j]) + Yimag[idxs[i], idxs[j]] * np.sin(Va[i] - Va[j]))
                jac[i + hid, j] = Vm[i] * Vm[j] * (-Yreal[idxs[i], idxs[j]] * np.cos(Va[i] - Va[j]) - Yimag[idxs[i], idxs[j]] * np.sin(Va[i] - Va[j]))
            else:
                sum_a += 0

        for j in slackids:
            sum_a += Vm[i] * (Yreal[idxs[i], j] * np.cos(Va[i]) + Yimag[idxs[i], j] * np.sin(Va[i]))
        jac[i+hid, i] = sum_a

        sum_v = 0
        # F by Vm[i]
        for j in range(hid):
            if i != j:
                sum_v += Vm[j] * (Yreal[idxs[i], idxs[j]] * np.sin(Va[i] - Va[j]) - Yimag[idxs[i], idxs[j]] * np.cos(Va[i] - Va[j]))
                jac[i+hid, j+hid] = Vm[i] * (Yreal[idxs[i], idxs[j]] * np.sin(Va[i] - Va[j]) - Yimag[idxs[i], idxs[j]] * np.cos(Va[i] - Va[j]))
            else:
                sum_v += -2*Vm[i] * Yimag[idxs[i], idxs[j]]

        for j in slackids:
            sum_v += (Yreal[idxs[i], j] * np.sin(Va[i]) - Yimag[idxs[i], j] * np.cos(Va[i]))
        jac[i+hid, i+hid] = sum_v

    return jac


def _scipypf(Ybus, Sbus, V0, pv, pq, options):
    """
        It's a modified newtonpf.py from pandapower package
    """

    # options
    tol = options['tolerance_mva']

    # set up indexing for updating V
    pvpq = r_[pv, pq]

    slackids = [x for x in range(len(V0)) if x not in pvpq]
    # make json for evaluation of F(x)
    tojson = dict()
    Ybus = Ybus.todense()
    tojson['Yreal'] = Ybus.real.tolist()
    tojson['Yimag'] = Ybus.imag.tolist()
    tojson['Sreal'] = Sbus.real.tolist()
    tojson['Simag'] = Sbus.imag.tolist()
    tojson['slackids'] = slackids
    tojson['pvpq'] = pvpq.tolist()
    tojson['algorithm'] = options["algorithm"]
    json.dump(tojson, open('fxjson.json', 'w'))

    Va = np.array(angle(V0).tolist())
    Vm = np.array(abs(V0).tolist())
    Va = Va[pvpq].copy()
    Vm = Vm[pvpq].copy()

    V = np.concatenate((Va, Vm))
    converged = True
    sol = []
    if options["algorithm"] == 'krylov':
        out = optimize.root(Fx, V, method='krylov', options={"jac_options": {"method": 'lgmres'}})
        converged = out.success
        sol = out.x
    if options["algorithm"] == 'anderson':
        out = optimize.root(Fx, V, jac=fxjac, method='anderson')
        converged = out.success
        sol = out.x
    if options["algorithm"] == 'broyden1':
        out = optimize.root(Fx, V, jac=fxjac, method='broyden1')
        converged = out.success
        sol = out.x
    if options["algorithm"] == 'broyden2':
        out = optimize.root(Fx, V, jac=fxjac, method='broyden2')
        converged = out.success
        sol = out.x
    if options["algorithm"] == 'hybr':
        out = optimize.root(Fx, V, jac=fxjac, method='hybr')
        converged = out.success
        sol = out.x
    if options["algorithm"] == 'lm':
        out = optimize.root(Fx, V, method='lm')
        converged = out.success
        sol = out.x
    if options["algorithm"] == 'quasi':
        sol, converged = quasi_fun(V0, Ybus, Sbus, pvpq, slackids, tol)
        tmp = np.concatenate((sol[len(Va):], sol[:len(Va)]))
        sol = tmp.copy()
    if options["algorithm"] == 'cont_newton':
        sol, converged = cont_newton(V, tol)

    V = sol[len(Va):] * exp(1j * sol[:len(Va)])
    V0[pvpq] = V

    its = 1
    J = None
    Vm_it = None
    Va_it = None

    return V0, converged, its, J, Vm_it, Va_it


# ----------------------------------------------------------------------------------------------------------------------
# Quasilinearisation method by M. H. Nagrial and H. M. Soliman, see Nagrial, M. H., & Soliman, H. M. (1984). Power flow solution using the modified quasilinearization method. Computers & Electrical Engineering, 11(4), 213–217. https://doi.org/10.1016/0045-7906(84)90025-9
# ----------------------------------------------------------------------------------------------------------------------
def quasi_fun(V, Y, S, pvpq, slackids, tol):
    e = V[pvpq].real
    v = V[pvpq].imag
    p = S[pvpq].real
    q = S[pvpq].imag
    G = Y.real
    B = Y.imag
    g0 = np.zeros((len(e), 1))
    b0 = np.zeros((len(e), 1))
    for i in slackids:
        g0 = g0 + G[pvpq, i]
        b0 = b0 + B[pvpq, i]
    B = B[pvpq, :][:, pvpq]
    G = G[pvpq, :][:, pvpq]

    x = np.concatenate((e, v))

    Y = np.block([[G, -B], [B, G]])
    y0 = np.vstack((g0, b0))
    y0 = np.array(y0.T)
    y0 = y0[0].copy()
    converged = False
    hid = len(e)
    maxit = 200
    it = 0
    while not converged and maxit > it:
        e = x[:hid].copy()
        v = x[hid:].copy()
        f = np.zeros(2*len(e))
        for i in range(len(e)):
            f[i] = -(p[i]*e[i]+q[i]*v[i])/(e[i]**2+v[i]**2)
        for i in range(len(e)):
            f[i+len(e)] = -(p[i] * v[i] - q[i] * e[i]) / (e[i] ** 2 + v[i] ** 2)
        r = np.array(np.dot(Y, x))[0] + y0 + f

        fp = np.dot(r, r)
        print('Error of quasi: ' + str(fp))
        if np.dot(r, r) < tol:
            converged = True
            break
        if np.dot(r, r) > 2.0:
            converged = False
            break
        D = np.zeros((len(e), len(e)))
        D1 = np.zeros((len(e), len(e)))
        for i in range(len(e)):
            D[i, i] = (p[i]*(e[i]**2-v[i]**1)+2*q[i]*e[i]*v[i])/((e[i]**2+v[i]**2)**2)
            D1[i, i] = (q[i]*(v[i]**2-e[i]**2)+2*p[i]*v[i]*e[i])/((e[i]**2+v[i]**2)**2)
        drdx = Y.T + np.block([[D, D1], [D1, D]])
        a = solve(drdx.T, -r)

        alpha = bisect(F_fun, 0, 1, x, Y, y0, a, p, q)
        if alpha is None:
            converged = True
            break
        x = x + alpha * a
        it = it + 1
    return x, converged


def F_fun(alpha, x, Y, y0, a, p, q):
    x = (x + alpha * a)
    hid = int(len(x)/2)
    e = x[:hid].copy()
    v = x[hid:].copy()
    f = np.zeros(2 * len(e))
    for i in range(len(e)):
        f[i] = -(p[i] * e[i] + q[i] * v[i]) / (e[i] ** 2 + v[i] ** 2)
    for i in range(len(e)):
        f[i + len(e)] = -(p[i] * v[i] - q[i] * e[i]) / (e[i] ** 2 + v[i] ** 2)
    r = np.array(np.dot(Y, x))[0] + y0 + f
    F = np.dot(r, r)
    return F


def bisect(F_fun, a, b, x, Y, y0, a_vec, p, q):
    if F_fun(1, x, Y, y0, a_vec, p, q) < F_fun(0, x, Y, y0, a_vec, p, q):
        return 1
    else:
        m = (a+b)/2

        if F_fun(m, x, Y, y0, a_vec, p, q) < F_fun(0, x, Y, y0, a_vec, p, q):
            return m
        else:
            ml = (a+m)/2
            mr = (b+m)/2
            if F_fun(ml, x, Y, y0, a_vec, p, q) < F_fun(m, x, Y, y0, a_vec, p, q):
                return bisect(F_fun, a, m, x, Y, y0, a_vec, p, q)
            elif F_fun(mr, x, Y, y0, a_vec, p, q) < F_fun(m, x, Y, y0, a_vec, p, q):
                return bisect(F_fun, m, b, x, Y, y0, a_vec, p, q)


# ----------------------------------------------------------------------------------------------------------------------
# Continuous Newtons method
# ----------------------------------------------------------------------------------------------------------------------
def cont_newton(V, tol):
    dt = 1
    err = np.linalg.norm(Fx(V), Inf)
    print('Error of cont_newton: ' + str(err))

    while err > tol:
        k1 = -np.dot(np.linalg.inv(fxjac(V)), Fx(V))
        k2 = -np.dot(np.linalg.inv(fxjac(V + 0.5*dt*k1)), Fx(V + 0.5*dt*k1))
        k3 = -np.dot(np.linalg.inv(fxjac(V + 0.5*dt*k2)), Fx(V + 0.5*dt*k2))
        k4 = -np.dot(np.linalg.inv(fxjac(V + dt*k3)), Fx(V + dt*k3))
        V = V + ((dt*(k1+2*k2+2*k3+k4))/6)

        eps = np.max(abs(k2.T[0]-V))
        if eps > 0.01:
            dt = max([0.985*dt, 0.75])
        else:
            dt = min([1.015 * dt, 0.75])
        err = np.linalg.norm(Fx(V), Inf)
        print('Error of cont_newton: ' + str(err))
    return V, True
